import inspect
import os

TEMPLATES_DIR = os.path.join(
    os.path.dirname(
        os.path.abspath(inspect.getframeinfo(inspect.currentframe()).filename)
    ),
    "templates",
)

_VALID_MATCHES = list(filter(lambda x: x.endswith(".rst"), os.listdir(TEMPLATES_DIR)))


def load(name):
    if not name.endswith(".rst"):
        name += ".rst"

    if name not in _VALID_MATCHES:
        raise ValueError("Unrecognised template: %s" % name)

    with open(os.path.join(TEMPLATES_DIR, name)) as f:
        contents = f.read()

    return contents


def _tab_indented_object_block(objects):
    block = "\n\t".join(objects)
    # return "\t" + block
    return block


def _header_break(name):
    return "".join(["=" for _ in range(len("``" + name + "``"))])


def make_subpkg_overview(name):
    header_break = _header_break(name)
    return load("subpkg_overview").format(header_break=header_break, name=name)


def make_classes_overview(dirname, classes):
    classes_block = _tab_indented_object_block(classes)
    return load("classes_overview").format(dirname=dirname, classes=classes_block)


def make_single_class_overview(name):
    return load("single_class_overview").format(name=name)


def make_functions_section(functions):
    functions_block = _tab_indented_object_block(functions)
    return load("functions_section").format(functions=functions_block)


def make_properties_section(properties):
    properties_block = _tab_indented_object_block(properties)
    return load("properties_section").format(properties=properties_block)


def make_methods_section(methods, constructor=None):
    methods_block = _tab_indented_object_block(methods)

    if constructor is not None:
        constructor_block = "\n\t%s" % constructor
    else:
        constructor_block = ""

    return load("methods_section").format(
        constructor=constructor_block, methods=methods_block
    )


def make_simplemod_header(
    name, modname, cls_name, cls_props, cls_methods, cls_init, functions
):
    header_break = _header_break(name)

    if cls_name is not None:
        class_overview = make_single_class_overview(cls_name)
    else:
        class_overview = ""

    if cls_props:
        prop_section = make_properties_section(cls_props)
    else:
        prop_section = ""

    if cls_methods:
        if cls_init is not None:
            constructor = "%s.%s" % (cls_name, cls_init)
        else:
            constructor = None

        meth_section = make_methods_section(cls_methods, constructor)
    else:
        meth_section = ""

    functions_section = make_functions_section(functions)

    return load("simplemod_header").format(
        header_break=header_break,
        name=name,
        module_name=modname,
        class_overview=class_overview,
        class_properties=prop_section,
        class_methods=meth_section,
        functions=functions_section,
    )


def make_singleclass_header(name, modname, cls_name, cls_props, cls_methods, cls_init):
    header_break = _header_break(name)

    class_overview = make_single_class_overview(cls_name)

    if cls_props:
        prop_section = make_properties_section(cls_props)
    else:
        prop_section = ""

    if cls_methods:
        if cls_init is not None:
            constructor = "%s.%s" % (cls_name, cls_init)
        else:
            constructor = None

        meth_section = make_methods_section(cls_methods, constructor)
    else:
        meth_section = ""

    return load("singleclass_header").format(
        header_break=header_break,
        name=name,
        module_name=modname,
        class_overview=class_overview,
        class_properties=prop_section,
        class_methods=meth_section,
    )


def make_multiclass_header(name, modname, classes_dirname, classes, functions):
    header_break = _header_break(name)

    classes_overview = make_classes_overview(classes_dirname, classes)

    functions_section = make_functions_section(functions)

    return load("multiclass_header").format(
        header_break=header_break,
        name=name,
        module_name=modname,
        classes_overview=classes_overview,
        functions=functions_section,
    )



import os
import re
import sys

from .parsing import SourceFileInfo
from . import templates


def write(subpkg_doc_files, submod_doc_files):
    for subpkg_doc_file in subpkg_doc_files:
        if os.path.isfile(subpkg_doc_file):
            continue

        name = subpkg_doc_file.split(os.sep)[-1].replace(".rst", "")
        contents = templates.load.make_subpkg_overview(name)

        with open(subpkg_doc_file, "w+") as f:
            f.write(_format_writable(contents))

    for sfi, doc_file in submod_doc_files.items():
        _manipulate_doc_file(doc_file, sfi)


def _format_writable(contents: str):
    contents = contents.strip()
    return re.sub(r"\n\s*\n", "\n\n", contents) + "\n"


def _manipulate_doc_file(doc_file, sfi: SourceFileInfo):
    # recursively make all directories up to doc_file if they don't exist already
    os.makedirs(os.path.dirname(doc_file), exist_ok=True)

    public_classes = sfi.public("classes")
    if len(public_classes) > 1:
        _multiclass_handler(doc_file, sfi, public_classes)
    else:
        if os.path.isfile(doc_file):
            # TODO complicated stuff
            return

        _, fname = os.path.split(doc_file)
        name = fname.replace(".rst", "")
        if not public_classes:
            cls_name = None
            cls_props = []
            cls_meths = []
            cls_init = None
        else:
            cls_name = list(public_classes.keys())[0]
            class_data = sfi.get_class_data(cls_name)

            cls_init = (
                None if class_data.constructor is None else class_data.constructor[0]
            )
            cls_props = list(class_data.properties.keys())
            cls_meths = list(class_data.methods.keys())

        contents = templates.load.make_simplemod_header(
            name,
            name,
            cls_name,
            cls_props,
            cls_meths,
            cls_init,
            list(sfi.functions.keys()),
        )

        with open(doc_file, "w+") as f:
            f.write(_format_writable(contents))


def _multiclass_handler(doc_file, sfi, classes):
    leading_dir, fname = os.path.split(doc_file)
    new_dirname = fname.split(".")[-2]

    full_new_dirname = os.path.join(leading_dir, new_dirname)
    if not os.path.isdir(full_new_dirname):
        os.mkdir(full_new_dirname)

    for pclass in classes:
        class_data = sfi.get_class_data(pclass)

        new_fname = ".".join(fname.split(".")[:-1]) + (".%s.rst" % pclass)
        fpath = os.path.join(leading_dir, os.path.join(new_dirname, new_fname))

        if os.path.isfile(fpath):
            # TODO complicated stuff
            continue

        with open(fpath, "w+") as f:
            name = new_fname.replace(".rst", "")
            mod_name = fname.replace(".rst", "")
            # FIXME: some issues here with properties / methods duplication
            cd_constructor = (
                None if class_data.constructor is None else class_data.constructor[0]
            )
            contents = templates.load.make_singleclass_header(
                name,
                mod_name,
                class_data.name,
                list(class_data.properties.keys()),
                list(class_data.methods.keys()),
                cd_constructor,
            )

            f.write(_format_writable(contents))

    if os.path.isfile(doc_file):
        # TODO complicated stuff
        return

    with open(doc_file, "w+") as f:
        name = fname.replace(".rst", "")
        contents = templates.load.make_multiclass_header(
            name, name, new_dirname, list(classes.keys()), list(sfi.functions.keys())
        )

        f.write(_format_writable(contents))
